@@ -6,12 +6,13 @@
     FLINT is free software: you can redistribute it and/or modify it under
     the terms of the GNU Lesser General Public License (LGPL) as published
     by the Free Software Foundation; either version 2.1 of the License, or
-    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
+    (at your option) any later version.  See <https://www.gnu.org/licenses/>.
 */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <gmp.h>
+#include <limits.h>
 
 #ifndef GMP_COMPAT_H
 #define GMP_COMPAT_H
@@ -29,7 +30,7 @@
   } while (0)
 
 
-#if defined(__MINGW64__) && !defined(__MPIR_VERSION)
+#if WORD_MAX != LONG_MAX && !defined(__MPIR_VERSION)
 
 #define FLINT_MOCK_MPZ_UI(xxx, yyy) \
    __mpz_struct (xxx)[1] = {{ 1, 0, NULL }}; \
@@ -57,13 +58,14 @@
 static __inline__
 void flint_mpz_set_si(mpz_ptr r, slong s)
 {
-#if __GNU_MP_RELEASE >= 60200
+   /* GMP 6.2 lazily performs allocation, deal with that if necessary
+      (in older GMP versions, this code is simply never triggered) */
    if (r->_mp_alloc == 0)
    {
       r->_mp_d = (mp_ptr) flint_malloc(sizeof(mp_limb_t));
       r->_mp_alloc = 1;
    }
-#endif
+
    if (s < 0) {
       r->_mp_size = -1;
       r->_mp_d[0] = -s;
@@ -76,13 +78,14 @@ void flint_mpz_set_si(mpz_ptr r, slong s)
 static __inline__
 void flint_mpz_set_ui(mpz_ptr r, ulong u)
 {
-#if __GNU_MP_RELEASE >= 60200
+   /* GMP 6.2 lazily performs allocation, deal with that if necessary
+      (in older GMP versions, this code is simply never triggered) */
    if (r->_mp_alloc == 0)
    {
       r->_mp_d = (mp_ptr) flint_malloc(sizeof(mp_limb_t));
       r->_mp_alloc = 1;
    }
-#endif
+
    r->_mp_d[0] = u; 
    r->_mp_size = u != 0;
 }
@@ -731,55 +734,6 @@ int flint_mpf_fits_slong_p(mpf_srcptr f)
   return fl <= (fs >= 0 ? (mp_limb_t) WORD_MAX : - (mp_limb_t) WORD_MIN);
 }
 
-/* double mpf_get_d_2exp (signed long int *exp, mpf_t src).
-
-Copyright 2001, 2002, 2003, 2004 Free Software Foundation, Inc.
-
-This file is part of the GNU MP Library.
-
-The GNU MP Library is free software; you can redistribute it and/or modify
-it under the terms of the GNU Lesser General Public License as published by
-the Free Software Foundation; either version 2.1 of the License, or (at your
-option) any later version.
-
-The GNU MP Library is distributed in the hope that it will be useful, but
-WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
-License for more details.
-
-You should have received a copy of the GNU Lesser General Public License
-along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
-the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-MA 02110-1301, USA. */
-
-extern double __gmpn_get_d(mp_limb_t *, size_t, size_t, long);
-
-static __inline__
-double flint_mpf_get_d_2exp(slong * exp2, mpf_srcptr src)
-{
-  mp_size_t size, abs_size;
-  mp_limb_t * ptr;
-  int cnt;
-  slong exp;
-
-  size = src->_mp_size;
-  if (size == 0)
-    {
-      *exp2 = 0;
-      return 0.0;
-    }
-
-  ptr = src->_mp_d;
-  abs_size = FLINT_ABS(size);
-  count_leading_zeros (cnt, ptr[abs_size - 1]);
-
-  exp = src->_mp_exp * FLINT_BITS - cnt;
-  *exp2 = exp;
-
-  return __gmpn_get_d (ptr, abs_size, size,
-                    (long) - (abs_size * FLINT_BITS - cnt));
-}
-
 #else
 
 #define flint_mpz_get_si mpz_get_si
@@ -829,8 +783,69 @@ double flint_mpf_get_d_2exp(slong * exp2, mpf_srcptr src)
 #define flint_mpf_get_si mpf_get_si
 #define flint_mpf_cmp_ui mpf_cmp_ui
 #define flint_mpf_fits_slong_p mpf_fits_slong_p
+
+#endif
+
+#if WORD_MAX == LONG_MAX
+
 #define flint_mpf_get_d_2exp mpf_get_d_2exp
 
+#else
+
+/* double mpf_get_d_2exp (signed long int *exp, mpf_t src).
+
+Copyright 2001, 2002, 2003, 2004 Free Software Foundation, Inc.
+
+This file is part of the GNU MP Library.
+
+The GNU MP Library is free software; you can redistribute it and/or modify
+it under the terms of the GNU Lesser General Public License as published by
+the Free Software Foundation; either version 2.1 of the License, or (at your
+option) any later version.
+
+The GNU MP Library is distributed in the hope that it will be useful, but
+WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+License for more details.
+
+You should have received a copy of the GNU Lesser General Public License
+along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
+the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
+MA 02110-1301, USA. */
+
+extern double __gmpn_get_d(mp_limb_t *, size_t, size_t, long);
+
+/*
+  This function is only required for MPIR >= 3.0.0 and for GMP, but we include
+  unconditionally when sizeof(ulong) != sizeof(unsigned long)
+*/
+
+static __inline__
+double flint_mpf_get_d_2exp(slong * exp2, mpf_srcptr src)
+{
+  mp_size_t size, abs_size;
+  mp_limb_t * ptr;
+  int cnt;
+  slong exp;
+
+  size = src->_mp_size;
+  if (size == 0)
+    {
+      *exp2 = 0;
+      return 0.0;
+    }
+
+  ptr = src->_mp_d;
+  abs_size = FLINT_ABS(size);
+  count_leading_zeros (cnt, ptr[abs_size - 1]);
+
+  exp = src->_mp_exp * FLINT_BITS - cnt;
+  *exp2 = exp;
+
+  return __gmpn_get_d (ptr, abs_size, size,
+                    (long) - (abs_size * FLINT_BITS - cnt));
+}
+
 #endif
 
 #endif
